﻿#include "renderer_text.h"

#include <future>
#include <utility>
#include <spdlog/spdlog.h>

#include "renderer.h"
#include "renderer_texture_array.h"

#define STB_TRUETYPE_IMPLEMENTATION
#include <stb_truetype.h>

#include "misc/scope_exit.h"

// 在SDFFontCache.cpp中添加实现
const std::wstring aorii_text::COMMON_ASCII =
    L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";

const std::wstring aorii_text::COMMON_CHINESE =
    L"的一是在不了有和人这中大为上个国我以要他时来用们生到作地于出就分对成会可主发年动同工也能下过子说产种面而方后多定行学法所民得经十三之进着等部度家电力里如水化高自二理起小物现实加量都两体制机当使点从业本去把性好应开它合还因由其些然前外天政四日那社义事平形相全表间样与关各重新线内数正心反你明看原又么利比或但质气第向道命此变条只没结解问意建月公无系军很情者最立代想已通并提直题党程展五果料象员革位入常文总次品式活设及管特件长求老头基资边流路级少图山统接知较将组见计别她手角期根论运农指几九区强放决西被干做必战先回则任取据处队南给色光门即保治北造百规热领七海口东导器压志世金增争济阶油思术极交受联什认六共权收证改清己美再采转更单风切打白教速花带安场身车例真务具万每目至达走积示议声报斗完类八离华名确才科张信马节话米整空元况今集温传土许步群广石记需段研界拉林律叫且究观越织装影算低持音众书布复容儿须际商非验连断深难近矿千周委素技备半办青省列习响约支般史感劳便团往酸历市克何除消构府称太准精值号率族维划选标写存候毛亲快效斯院查江型眼王按格养易置派层片始却专状育厂京识适属圆包火住调满县局照参红细引听该铁价严";

const std::wstring aorii_text::COMMON_PUNCTUATION =
    L"，。！？：；''""《》【】（）、";

const std::wstring aorii_text::COMMON_NUMBERS =
    L"0123456789";

aorii_text::aorii_text(const uint32_t in_texture_size) : font(nullptr), texture_array(nullptr),
    texture_size(in_texture_size),
    current_x(0), current_y(0), current_texture_index(0), scale(0) {
}

aorii_text::~aorii_text() {
    delete font;
}

bool aorii_text::init_freetype() {
    return true;
}

void aorii_text::destroy_freetype() {
}

bool aorii_text::initialize(const wchar_t* in_font_path, const float in_font_pixel_size) {
    const std::string font_path_a(in_font_path, in_font_path + wcslen(in_font_path));

    font = new stbtt_fontinfo();
    font_pixel_size = in_font_pixel_size;

    // 加载字体文件
    FILE* font_file = nullptr;
    fopen_s(&font_file, font_path_a.c_str(), "rb");
    if (!font_file) {
        printf("Failed to open font file\n");
        return false;
    }

    fseek(font_file, 0, SEEK_END);
    const size_t font_size = ftell(font_file);
    fseek(font_file, 0, SEEK_SET);

    auto* font_buffer = static_cast<uint8_t*>(malloc(font_size));
    fread(font_buffer, 1, font_size, font_file);
    fclose(font_file);

    if (!stbtt_InitFont(font, font_buffer, stbtt_GetFontOffsetForIndex(font_buffer, 0))) {
        spdlog::error("Failed to initialize font");
        free(font_buffer);
        delete font;
        font = nullptr;
        return false;
    }

    scale = stbtt_ScaleForPixelHeight(font, in_font_pixel_size);

    // 获取字体的垂直度量
    stbtt_GetFontVMetrics(font, &ascent, &descent, &line_gap);
    ascent *= scale;
    descent *= scale;
    line_gap *= scale;
    spdlog::info("Font vertical metrics:\n  上升距离: {}, 下降距离: {}, 行间距: {}", ascent, descent, line_gap);

    // Create initial texture atlas
    return create_new_texture_atlas();
}

float aorii_text::get_kerning(wchar_t ch1, wchar_t ch2) const {
    // 获取字距调整
    return stbtt_GetCodepointKernAdvance(font, ch1, ch2) * scale;
}

bool aorii_text::create_new_texture_atlas() {
    // 预计需要的纹理数量(8张纹理几乎可以容纳所有字符)
    constexpr uint32_t expected_textures = 8;
    texture_array = aorii::get_renderer_raw()->create_texture_array(texture_size, texture_size, expected_textures, texture_format::R8_UNORM, device_memory_type::gpu);
    return texture_array != nullptr;
}

bool aorii_text::get_or_create_character(const wchar_t ch, character_info& out_info) {
    if (const auto it = character_map.find(ch); it != character_map.end()) {
        out_info = it->second;
        return true;
    }

    // 加载目标字符（以 'A' 为例）
    int glyph_index = stbtt_FindGlyphIndex(font, ch);
    if (glyph_index == 0) {
        printf("Failed to find glyph for character\n");
        return false;
    }

    // Try to add to current atlas
    out_info = add_character_to_atlas(glyph_index, ch);
    return true;
}

bool aorii_text::precache_characters(const std::wstring& characters) {
    if (!font) {
        spdlog::error("Font not initialized");
        return false;
    }

    // 创建进度通知回调的函数指针类型
    using ProgressCallback = std::function<void(float progress, const wchar_t currentChar)>;

    bool success = true;

    // 使用线程池并行生成SDF数据
    const size_t max_threads = std::thread::hardware_concurrency();
    std::vector<std::future<std::vector<uint8_t>>> futures;
    std::mutex map_mutex;

    for (size_t i = 0; i < characters.length(); i++) {
        wchar_t ch = characters[i];

        // 检查字符是否已经缓存
        if (character_map.contains(ch)) {
            continue;
        }

        // 加载目标字符（以 'A' 为例）
        int glyph_index = stbtt_FindGlyphIndex(font, ch);
        if (glyph_index == 0) {
            continue;
        }

        // 生成SDF数据
        add_character_to_atlas(glyph_index, ch);
    }

    return success;
}

bool aorii_text::precache_common_characters() {
    std::wstring all_common_chars;
    all_common_chars.reserve(COMMON_ASCII.length() +
                          COMMON_CHINESE.length() +
                          COMMON_PUNCTUATION.length() +
                          COMMON_NUMBERS.length());

    all_common_chars += COMMON_ASCII;
    all_common_chars += COMMON_CHINESE;
    all_common_chars += COMMON_PUNCTUATION;
    all_common_chars += COMMON_NUMBERS;

    return precache_characters(all_common_chars);
}

uint32_t aorii_text::get_font_char_count() const { return font->numGlyphs; }

character_info& aorii_text::add_character_to_atlas(int32_t glyph_index, const wchar_t ch) {
    // 获取SDF尺寸和位图
    int32_t width, height, x_offset, y_offset;
    constexpr int padding = 8;
    constexpr uint8_t on_edge_value = 128;
    constexpr float pixel_dist_scale = 16.f;

    auto* sdf_bitmap = stbtt_GetGlyphSDF(
        font,
        scale,
        glyph_index,
        padding,
        on_edge_value,
        pixel_dist_scale,
        &width, &height,
        &x_offset, &y_offset
    );

    if (!sdf_bitmap) {
        throw std::runtime_error("Failed to generate SDF for glyph");
    }
    ON_SCOPE_EXIT {
        stbtt_FreeSDF(sdf_bitmap, font->userdata);
    };

    // 检查当前行是否有足够空间
    if (current_x + width > texture_size) {
        // 移动到下一行
        current_x = 0;
        current_y = next_row_y;
    }

    // 检查是否需要新的纹理
    if (current_y + height > texture_size) {
        current_texture_index++;
        if (current_texture_index >= texture_array->get_count()) {
            throw std::runtime_error("Texture array capacity exceeded");
        }
        current_x = 0;
        current_y = 0;
        next_row_y = 0;
    }

    // 更新纹理
    Eigen::AlignedBox2i box;
    box.min() = Eigen::Vector2i(current_x, current_y);
    box.max() = Eigen::Vector2i(current_x + width, current_y + height);

    texture_array->update_subresource(current_texture_index, box, sdf_bitmap, width);

    // 获取字形度量信息
    int advance_width, left_side_bearing;
    stbtt_GetGlyphHMetrics(font, glyph_index, &advance_width, &left_side_bearing);

    // 创建并存储字符信息
    character_info info;
    info.tex_u = current_x / static_cast<float>(texture_size);
    info.tex_v = current_y / static_cast<float>(texture_size);
    info.advance = advance_width * scale;
    info.tex_z = static_cast<float>(current_texture_index);
    info.width = width;
    info.height = height;
    info.x_offset = x_offset;
    info.y_offset = y_offset;

    // 更新位置和下一行的y坐标
    current_x += width;
    next_row_y = std::max(next_row_y, current_y + height);

    return character_map[ch] = info;
}

